#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class Filter(FunctionNode):
    @staticmethod
    def date(start: str, end: str = None) -> engine.Filter:
        if start is not None and not isinstance(start, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"start 只支持str类型参数, 传入类型为{type(start)}"
            )

        if end is not None and not isinstance(end, str):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"end 只支持str类型参数, 传入类型为{type(end)}")

        invoke_args = {
            "start": start,
            "end": end,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "start" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数start不能为空")

        return FunctionHelper.apply("Filter.date", "engine.Filter", invoke_args)

    @staticmethod
    def bounds(
        geometry: Union[engine.Geometry, engine.Feature, engine.FeatureCollection]
    ) -> engine.Filter:
        if geometry is not None and not isinstance(
            geometry, (engine.Geometry, engine.Feature, engine.FeatureCollection)
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"geometry 只支持(engine.Geometry,engine.Feature,engine.FeatureCollection)类型参数, 传入类型为{type(geometry)}",
            )

        invoke_args = {
            "geometry": geometry,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "geometry" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数geometry不能为空")

        return FunctionHelper.apply("Filter.bounds", "engine.Filter", invoke_args)

    @staticmethod
    def eq(name: str, value: object) -> engine.Filter:
        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        if value is not None and not isinstance(value, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"value 只支持object类型参数, 传入类型为{type(value)}"
            )

        invoke_args = {
            "name": name,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "name" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数name不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Filter.eq", "engine.Filter", invoke_args)

    @staticmethod
    def gt(name: str, value: object) -> engine.Filter:
        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        if value is not None and not isinstance(value, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"value 只支持object类型参数, 传入类型为{type(value)}"
            )

        invoke_args = {
            "name": name,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "name" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数name不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Filter.gt", "engine.Filter", invoke_args)

    @staticmethod
    def gte(name: str, value: object) -> engine.Filter:
        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        if value is not None and not isinstance(value, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"value 只支持object类型参数, 传入类型为{type(value)}"
            )

        invoke_args = {
            "name": name,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "name" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数name不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Filter.gte", "engine.Filter", invoke_args)

    @staticmethod
    def lt(name: str, value: object) -> engine.Filter:
        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        if value is not None and not isinstance(value, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"value 只支持object类型参数, 传入类型为{type(value)}"
            )

        invoke_args = {
            "name": name,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "name" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数name不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Filter.lt", "engine.Filter", invoke_args)

    @staticmethod
    def lte(name: str, value: object) -> engine.Filter:
        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        if value is not None and not isinstance(value, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"value 只支持object类型参数, 传入类型为{type(value)}"
            )

        invoke_args = {
            "name": name,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "name" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数name不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Filter.lte", "engine.Filter", invoke_args)

    @staticmethod
    def neq(name: str, value: object) -> engine.Filter:
        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        if value is not None and not isinstance(value, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"value 只支持object类型参数, 传入类型为{type(value)}"
            )

        invoke_args = {
            "name": name,
            "value": value,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "name" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数name不能为空")

        if "value" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数value不能为空")

        return FunctionHelper.apply("Filter.neq", "engine.Filter", invoke_args)

    @staticmethod
    def And(filters: list) -> engine.Filter:
        if filters is not None and not isinstance(filters, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"filters 只支持list类型参数, 传入类型为{type(filters)}"
            )

        invoke_args = {
            "filters": filters,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "filters" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数filters不能为空")

        return FunctionHelper.apply("Filter.and", "engine.Filter", invoke_args)

    @staticmethod
    def Or(filters: list) -> engine.Filter:
        if filters is not None and not isinstance(filters, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"filters 只支持list类型参数, 传入类型为{type(filters)}"
            )

        invoke_args = {
            "filters": filters,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "filters" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数filters不能为空")

        return FunctionHelper.apply("Filter.or", "engine.Filter", invoke_args)

    @staticmethod
    def Not() -> engine.Filter:
        invoke_args = {}

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Filter.not", "engine.Filter", invoke_args)

    @staticmethod
    def calendarRange(
        start: int, end: int = None, field: str = "day_of_year"
    ) -> engine.Filter:
        if start is not None and not isinstance(start, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"start 只支持int类型参数, 传入类型为{type(start)}"
            )

        if end is not None and not isinstance(end, int):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"end 只支持int类型参数, 传入类型为{type(end)}")

        if field is not None and not isinstance(field, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"field 只支持str类型参数, 传入类型为{type(field)}"
            )

        invoke_args = {
            "start": start,
            "end": end,
            "field": field,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "start" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数start不能为空")

        return FunctionHelper.apply(
            "Filter.calendarRange", "engine.Filter", invoke_args
        )

    @staticmethod
    def dateRangeContains(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Filter.dateRangeContains", "engine.Filter", invoke_args
        )

    @staticmethod
    def dayOfYear(start: int, end: int) -> engine.Filter:
        if start is not None and not isinstance(start, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"start 只支持int类型参数, 传入类型为{type(start)}"
            )

        if end is not None and not isinstance(end, int):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"end 只支持int类型参数, 传入类型为{type(end)}")

        invoke_args = {
            "start": start,
            "end": end,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "start" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数start不能为空")

        if "end" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数end不能为空")

        return FunctionHelper.apply("Filter.dayOfYear", "engine.Filter", invoke_args)

    @staticmethod
    def equals(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Filter.equals", "engine.Filter", invoke_args)

    @staticmethod
    def greaterThan(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Filter.greaterThan", "engine.Filter", invoke_args)

    @staticmethod
    def greaterThanOrEquals(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Filter.greaterThanOrEquals", "engine.Filter", invoke_args
        )

    @staticmethod
    def lessThan(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Filter.lessThan", "engine.Filter", invoke_args)

    @staticmethod
    def lessThanOrEquals(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Filter.lessThanOrEquals", "engine.Filter", invoke_args
        )

    @staticmethod
    def notEquals(
        leftField: str = None,
        rightValue: object = None,
        rightField: str = None,
        leftValue: object = None,
    ) -> engine.Filter:
        if leftField is not None and not isinstance(leftField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"leftField 只支持str类型参数, 传入类型为{type(leftField)}"
            )

        if rightValue is not None and not isinstance(rightValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightValue 只支持object类型参数, 传入类型为{type(rightValue)}",
            )

        if rightField is not None and not isinstance(rightField, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"rightField 只支持str类型参数, 传入类型为{type(rightField)}",
            )

        if leftValue is not None and not isinstance(leftValue, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"leftValue 只支持object类型参数, 传入类型为{type(leftValue)}",
            )

        invoke_args = {
            "leftField": leftField,
            "rightValue": rightValue,
            "rightField": rightField,
            "leftValue": leftValue,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Filter.notEquals", "engine.Filter", invoke_args)

    @staticmethod
    def notNull(properties: list) -> engine.Filter:
        if properties is not None and not isinstance(properties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"properties 只支持list类型参数, 传入类型为{type(properties)}",
            )

        invoke_args = {
            "properties": properties,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "properties" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数properties不能为空")

        return FunctionHelper.apply("Filter.notNull", "engine.Filter", invoke_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import fcntl
import json
import os
import struct
import numpy as np
from aiearth.core.error import AIEError, AIEErrorCode

from aiearth.engine.udf_base import UDFImages_pb2
from aiearth.engine.udf_base.images import DataType, Band, Image, ImageSet


class UDFUtils(object):
    @staticmethod
    def tile_data_decode(dataType: DataType, original_data: bytes) -> np.ndarray:
        def handle_invalid():
            raise AIEError(AIEErrorCode.DEFAULT_INTERNAL_ERROR, "不支持的数据类型=" + dataType.value)

        '''
        np.int8：有符号8位整数类型；
        np.uint8：无符号8位整数类型；
        np.int16：有符号16位整数类型；
        np.uint16：无符号16位整数类型；
        np.int32：有符号32位整数类型；
        np.float32：单精度浮点数类型；
        np.float64：双精度浮点数类型；
        '''
        decode_switcher = {
            DataType.UINT8: lambda: np.frombuffer(original_data, dtype=np.uint8),
            DataType.INT8: lambda: np.frombuffer(original_data, dtype=np.int8),
            DataType.UINT16: lambda: np.frombuffer(original_data, dtype=np.uint16),
            DataType.INT16: lambda: np.frombuffer(original_data, dtype=np.uint16),
            DataType.INT32: lambda: np.frombuffer(original_data, dtype=np.int32),
            DataType.FLOAT32: lambda: np.frombuffer(original_data, dtype=np.float32),
            DataType.FLOAT64: lambda: np.frombuffer(original_data, dtype=np.float64),
        }
        return decode_switcher.get(dataType, handle_invalid)()

    @staticmethod
    def tile_data_encode(dataType: DataType, original_data: np.ndarray) -> bytes:
        def handle_invalid():
            raise AIEError(AIEErrorCode.ARGS_ERROR, "不支持的数据类型=" + dataType.value)

        '''
        'b'：有符号的8位整数类型；
        'B'：无符号的8位整数类型；
        'h'：有符号的16位整数类型；
        'H'：无符号的16位整数类型；
        'i'：有符号的32位整数类型；
        'f'：单精度浮点数类型；
        'd'：双精度浮点数类型；
        '''
        encode_switcher = {
            DataType.UINT8: lambda: struct.pack('<' + 'B' * len(original_data), *original_data),
            DataType.INT8: lambda: struct.pack('<' + 'b' * len(original_data), *original_data),
            DataType.UINT16: lambda: struct.pack('<' + 'H' * len(original_data), *original_data),
            DataType.INT16: lambda: struct.pack('<' + 'h' * len(original_data), *original_data),
            DataType.INT32: lambda: struct.pack('<' + 'i' * len(original_data), *original_data),
            DataType.FLOAT32: lambda: struct.pack('<' + 'f' * len(original_data), *original_data),
            DataType.FLOAT64: lambda: struct.pack('<' + 'd' * len(original_data), *original_data),
        }
        return encode_switcher.get(dataType, handle_invalid)()

    @staticmethod
    def images_wrapper_to_set(imagesWrapper) -> ImageSet:
        images_wrapper = UDFImages_pb2.Wrapper()
        images_wrapper.ParseFromString(imagesWrapper)
        image_list = []
        params = images_wrapper.params
        for img in images_wrapper.image:
            bands = {}
            band_data = img.bandData
            for band_name, band in band_data.items():
                data_type = DataType.get_type_by_value(band.dataType)
                decode_data = UDFUtils.tile_data_decode(data_type, band.tileData)
                nodata_value = band.nodataValue
                cols = band.cols
                rows = band.rows
                bands[band_name] = Band(decode_data, data_type, nodata_value, cols, rows)
            image_list.append(Image(bands))
        return ImageSet(image_list, json.loads(params))

    @staticmethod
    def image_to_wrapper(image: Image) -> UDFImages_pb2.Image:
        image_wrapper = UDFImages_pb2.Image()
        for band_name, band in image.bands.items():
            wrapper_band = UDFImages_pb2.Band()
            wrapper_band.tileData = UDFUtils.tile_data_encode(band.data_type, band.data_array)
            wrapper_band.dataType = band.data_type.value
            wrapper_band.nodataValue = str(band.nodata_value)
            wrapper_band.cols = band.cols
            wrapper_band.rows = band.rows
            image_wrapper.bandData[band_name].CopyFrom(wrapper_band)
        return image_wrapper

    @staticmethod
    def wrapper_to_image(image_bytes: bytes) -> Image:
        image_wrapper = UDFImages_pb2.Image()
        image_wrapper.ParseFromString(image_bytes)
        bands = {}
        band_data = image_wrapper.bandData
        for band_name, band in band_data.items():
            data_type = DataType.get_type_by_value(band.dataType)
            decode_data = UDFUtils.tile_data_decode(data_type, band.tileData)
            nodata_value = band.nodataValue
            cols = band.cols
            rows = band.rows
            bands[band_name] = Band(decode_data, data_type, nodata_value, cols, rows)
        return Image(bands)

    @staticmethod
    def save_file(file_path, content):
        if not os.path.exists(file_path):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, "w") as file:
                try:
                    fcntl.flock(file, fcntl.LOCK_EX)
                    file.write(content)
                finally:
                    fcntl.flock(file, fcntl.LOCK_UN)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
Image.load
ImageCollection.load
ImageCollection.fromImages
Feature
FeatureCollection.load
FeatureCollection.fromFeatures
'''

ImageConstructorsImpl = '''
    def __init__(self, args=None) -> engine.Image:
        if isinstance(args, (int, float, complex)):
            invoke_args = {"value": args}
            super(Image, self).__init__("Image.constant", invoke_args)
        elif isinstance(args, str):
            invoke_args = {"id": args}
            super(Image, self).__init__("Image.load", invoke_args)
        elif isinstance(args, (list, tuple)):
            images = [Image(i) for i in args]
            result = images[0]
            for image in images[1:]:
                invoke_args = {
                    "srcImg": image,
                    "dstImg": result
                }
                result = FunctionHelper.apply("Image.addBands", "engine.Image", invoke_args)
            super(Image, self).__init__(result.func_name, result.invoke_args, result.var_name)
        elif args is None:
            image = Image(0)
            invoke_args = {
                "input": image,
                "mask": image
            }
            super(Image, self).__init__("Image.mask", invoke_args)
        elif isinstance(args, engine.variable_node.VariableNode):
            super(Image, self).__init__(args.func_name, args.invoke_args, args.var_name)
        elif isinstance(args, engine.function_node.FunctionNode):
            super(Image, self).__init__(args.func_name, args.invoke_args, args.var_name)
        else:
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"args 只支持number|str|list类型参数, 传入类型为{type(args)}")
'''

ImageCollectionConstructorsImpl = '''
    def __init__(self, args) -> engine.ImageCollection:
        if isinstance(args, str):
            invoke_args = {"id": args}
            super(ImageCollection, self).__init__("ImageCollection.load", invoke_args)
        elif isinstance(args, engine.Image):
            args = [args]
            invoke_args = {"images": args}
            super(ImageCollection, self).__init__("ImageCollection.fromImages", invoke_args)
        elif isinstance(args, (list, tuple)):
            images = [engine.Image(i) for i in args]
            invoke_args = {"images": images}
            super(ImageCollection, self).__init__("ImageCollection.fromImages", invoke_args)
        elif isinstance(args, engine.List):
            super(ImageCollection, self).__init__(
                "ImageCollection.fromImages", {"images": args}
            )
        else:
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"args 只支持str|engine.Image|list|engine.List类型参数, 传入类型为{type(args)}")
'''

FeatureConstructorsImpl = '''
    def __init__(self, geometry, properties=None) -> engine.Feature:
        args = geometry
        if isinstance(args, engine.Geometry):
            if properties is not None and not isinstance(properties, (dict, engine.function_node.FunctionNode)):
                raise AIEError(AIEErrorCode.ARGS_ERROR, f"properties 只支持dict|engine.function_node.FunctionNode类型参数, 传入类型为{type(properties)}")

            invoke_args = {
                "geometry": geometry,
                "properties": properties,
            }
            
            invoke_args = {k: v for k, v in invoke_args.items() if v is not None}
            super(Feature, self).__init__("Feature", invoke_args)
        elif isinstance(args, engine.variable_node.VariableNode):
            super(Feature, self).__init__(args.func_name, args.invoke_args, args.var_name)
        elif isinstance(args, engine.function_node.FunctionNode):
            super(Feature, self).__init__(args.func_name, args.invoke_args, args.var_name)
        elif args is None:
            invoke_args = {
                "geometry": geometry,
                "properties": properties,
            }
            
            invoke_args = {k: v for k, v in invoke_args.items() if v is not None}
            super(Feature, self).__init__("Feature", invoke_args)
        else:
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"geometry 只支持engine.Geometry类型参数, 传入类型为{type(geometry)}")

'''

FeatureCollectionConstructorsImpl = '''
    def __init__(self, args) -> engine.FeatureCollection:
        if isinstance(args, str):
            invoke_args = {"id": args}
            super(FeatureCollection, self).__init__("FeatureCollection.load", invoke_args)
        elif isinstance(args, engine.Feature):
            args = [args]
            invoke_args = {"features": args}
            super(FeatureCollection, self).__init__("FeatureCollection.fromFeatures", invoke_args)
        elif isinstance(args, (list, tuple)):
            features = [engine.Feature(i) for i in args]
            invoke_args = {"features": features}
            super(FeatureCollection, self).__init__("FeatureCollection.fromFeatures", invoke_args)
        elif isinstance(args, engine.List):
            super(FeatureCollection, self).__init__(
                "FeatureCollection.fromFeatures", {"features": args}
            )
        else:
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"args 只支持str|engine.Feature|list|engine.List类型参数, 传入类型为{type(args)}")
'''

CollectionElementTypeImpl = '''
    @abc.abstractmethod
    def elementType(self):
        pass
'''
ImageCollectionElementTypeImpl = '''
    def elementType(self):
        return engine.Image
'''
FeatureCollectionElementTypeImpl = '''
    def elementType(self):
        return engine.Feature
'''

CollectionMapImpl = '''
    def map(self, baseAlgorithm) -> engine.Collection:
        import types
        if baseAlgorithm is not None and not isinstance(baseAlgorithm, types.FunctionType):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"baseAlgorithm 只支持function类型参数, 传入类型为{type(baseAlgorithm)}")

        args = inspect.getfullargspec(baseAlgorithm).args
        mapping_args = ["_MAPPING_VAR_#arg" + str(k) + "_" + v for k,v in enumerate(args)]
        variables = [VariableNode(mapping_arg) for mapping_arg in mapping_args]

        element_type = self.elementType()
        def func_warp(e): return baseAlgorithm(element_type(e))
        body = func_warp(*variables)
        if body is None:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "自定义map函数必须有返回值")
        customfunc = CustomFunctionNode(mapping_args, body)

        invoke_args = {
            "collection": self,
            "baseAlgorithm": customfunc,
        }
        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "baseAlgorithm" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数baseAlgorithm不能为空")

        return FunctionHelper.apply("Collection.map", "engine.Collection", invoke_args)
'''

CollectionIterateImpl = '''
    def iterate(self, func: FunctionType, first: object) -> object:
        if func is not None and not isinstance(func, FunctionType):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"func 只支持FunctionType类型参数, 传入类型为{type(func)}"
            )

        if first is not None and not isinstance(first, object):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"first 只支持object类型参数, 传入类型为{type(first)}"
            )

        args = inspect.getfullargspec(func).args
        mapping_args = [
            "_MAPPING_VAR_#arg" + str(k) + "_" + v for k, v in enumerate(args)
        ]
        variables = [VariableNode(mapping_arg) for mapping_arg in mapping_args]

        element_type = self.elementType()

        def func_warp(collection_element_self, param_object: object):
            return func(element_type(collection_element_self), param_object)

        from copy import deepcopy
        deepcopy_first = deepcopy(first)
        setattr(deepcopy_first, "iterables", variables[1].var_name)
        body = func_warp(variables[0], deepcopy_first)
        if body is None:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "自定义map函数必须有返回值")
        customfunc = CustomFunctionNode(mapping_args, body)

        invoke_args = {
            "collection": self,
            "baseAlgorithm": customfunc,
            "first": first,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "baseAlgorithm" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数func不能为空")

        if "first" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数first不能为空")

        return FunctionHelper.apply("Collection.iterate", "object", invoke_args)
'''

ImageCollectionMapImpl = """
    def map(self, baseAlgorithm: types.FunctionType) -> Union[engine.ImageCollection, engine.FeatureCollection]:

        node = super(ImageCollection, self).map(baseAlgorithm)
        baseAlgorithm_return_type = type(node.invoke_args["baseAlgorithm"].body)
        if engine.Feature == baseAlgorithm_return_type:
            return FunctionHelper.cast(node, "engine.FeatureCollection")
        else:
            return FunctionHelper.cast(node, "engine.ImageCollection")
"""

FeatureCollectionMapImpl = """
    def map(self, baseAlgorithm: types.FunctionType) -> Union[engine.ImageCollection, engine.FeatureCollection]:

        node = super(FeatureCollection, self).map(baseAlgorithm)
        baseAlgorithm_return_type = type(node.invoke_args["baseAlgorithm"].body)
        if engine.Image == baseAlgorithm_return_type:
            return FunctionHelper.cast(node, "engine.ImageCollection")
        else:
            return FunctionHelper.cast(node, "engine.FeatureCollection")
"""

GetMapIdImpl = '''
    def getMapId(self, vis_params):
        if vis_params is not None and not isinstance(vis_params, dict):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"vis_params 只支持dict类型参数, 传入类型为{type(vis_params)}")
        return engine.client.Maps.getMapId(self, vis_params)
'''

ImageSelectImpl = '''
    def select(self, bandSelectors: Union[str, list]) -> engine.Image:

        if bandSelectors is not None and not isinstance(bandSelectors, (str, list)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"bandSelectors 只支持(str,list)类型参数, 传入类型为{type(bandSelectors)}",
            )

        if isinstance(bandSelectors, str):
            bandSelectors = [bandSelectors]

        invoke_args = {
            "input": self,
            "bandSelectors": bandSelectors,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "bandSelectors" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数bandSelectors不能为空")

        return FunctionHelper.apply("Image.select", "engine.Image", invoke_args)
'''

ImageCollectionSelectImpl = '''
    def select(self, selectors) -> engine.ImageCollection:
        return self.map(lambda image: image.select(selectors))
'''

GetCenterImpl = '''
    def getCenter(self) -> tuple:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            center = ((bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2)
            return center
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Center失败. bbox: {bbox}")
'''

GetBoundsImpl = '''
    def getBounds(self) -> list:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            bounds = [bbox[0], bbox[1], bbox[2], bbox[3]]
            return bounds
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Bounds失败. bbox: {bbox}")
'''

ClassifierTrainImpl = '''
    def train(
        self, features: engine.FeatureCollection, classProperty: str, inputProperties: list
    ) -> engine.Classifier:

        if features is not None and not isinstance(features, engine.FeatureCollection):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"features 只支持engine.FeatureCollection类型参数, 传入类型为{type(features)}",
            )

        if classProperty is not None and not isinstance(classProperty, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"classProperty 只支持str类型参数, 传入类型为{type(classProperty)}",
            )

        if inputProperties is not None and not isinstance(inputProperties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"inputProperties 只支持list类型参数, 传入类型为{type(inputProperties)}",
            )

        invoke_args = {
            "input": self,
            "features": features,
            "classProperty": classProperty,
            "inputProperties": inputProperties,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "features" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数features不能为空")

        if "classProperty" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数classProperty不能为空")

        if "inputProperties" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数inputProperties不能为空")

        clf = FunctionHelper.apply("Classifier.train", "engine.Classifier", invoke_args)

        obj = engine.client.InteractiveSession.classifierTrain(clf)
        invoke_args['preResult'] = obj
        clf = FunctionHelper.apply("Classifier.train", "engine.Classifier", invoke_args)

        return clf
'''

GeometryConstructorsImpl = '''
    def __init__(self, geoJson: object) -> engine.Geometry:
        def _isValidGeometry(geometry):
            if not isinstance(geometry, dict):
                raise AIEError(AIEErrorCode.ARGS_ERROR, f"geoJson 不合法. 类型应该为dict")
            
            if "type" not in geometry:
                raise AIEError(AIEErrorCode.ARGS_ERROR, f"geoJson 不合法. 缺少type")

        _isValidGeometry(geoJson)

        geo_type = geoJson['type']
        coordinates = geoJson.get('coordinates')
        geometries = geoJson.get('geometries')

        ctor_args = {}
        if geo_type == 'GeometryCollection':
            ctor_name = 'MultiGeometry'
            ctor_args['geometries'] = [Geometry(g) for g in geometries]
        else:
            ctor_name = geo_type
            ctor_args['coordinates'] = coordinates

        super(Geometry, self).__init__("GeometryConstructors." + ctor_name, ctor_args)
'''


ModelConstructorsImpl = '''
    def __init__(self, modelName) -> engine.Model:
        self.modelName = modelName
        super(Model, self).__init__("Model.load" , {
            "modelName": modelName
        })
'''

IMPL = {k: v for k, v in vars().items() if not k.startswith("__")}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class Geometry(FunctionNode):
    def getCenter(self) -> tuple:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            center = ((bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2)
            return center
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Center失败. bbox: {bbox}")

    def getBounds(self) -> list:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            bounds = [bbox[0], bbox[1], bbox[2], bbox[3]]
            return bounds
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Bounds失败. bbox: {bbox}")

    def __init__(self, geoJson: object) -> engine.Geometry:
        def _isValidGeometry(geometry):
            if not isinstance(geometry, dict):
                raise AIEError(AIEErrorCode.ARGS_ERROR, f"geoJson 不合法. 类型应该为dict")

            if "type" not in geometry:
                raise AIEError(AIEErrorCode.ARGS_ERROR, f"geoJson 不合法. 缺少type")

        _isValidGeometry(geoJson)

        geo_type = geoJson["type"]
        coordinates = geoJson.get("coordinates")
        geometries = geoJson.get("geometries")

        ctor_args = {}
        if geo_type == "GeometryCollection":
            ctor_name = "MultiGeometry"
            ctor_args["geometries"] = [Geometry(g) for g in geometries]
        else:
            ctor_name = geo_type
            ctor_args["coordinates"] = coordinates

        super(Geometry, self).__init__("GeometryConstructors." + ctor_name, ctor_args)

    def getMapId(self, vis_params):
        if vis_params is not None and not isinstance(vis_params, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"vis_params 只支持dict类型参数, 传入类型为{type(vis_params)}",
            )
        return engine.client.Maps.getMapId(self, vis_params)

    @staticmethod
    def BBox(
        west: [int, float], south: [int, float], east: [int, float], north: [int, float]
    ) -> engine.Geometry:
        if west is not None and not isinstance(west, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"west 只支持(int,float)类型参数, 传入类型为{type(west)}"
            )

        if south is not None and not isinstance(south, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"south 只支持(int,float)类型参数, 传入类型为{type(south)}"
            )

        if east is not None and not isinstance(east, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"east 只支持(int,float)类型参数, 传入类型为{type(east)}"
            )

        if north is not None and not isinstance(north, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"north 只支持(int,float)类型参数, 传入类型为{type(north)}"
            )

        invoke_args = {
            "west": west,
            "south": south,
            "east": east,
            "north": north,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "west" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数west不能为空")

        if "south" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数south不能为空")

        if "east" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数east不能为空")

        if "north" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数north不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.BBox", "engine.Geometry", invoke_args
        )

    @staticmethod
    def Point(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.Point", "engine.Geometry", invoke_args
        )

    @staticmethod
    def Polygon(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.Polygon", "engine.Geometry", invoke_args
        )

    @staticmethod
    def Rectangle(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.Rectangle", "engine.Geometry", invoke_args
        )

    @staticmethod
    def LineString(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.LineString", "engine.Geometry", invoke_args
        )

    @staticmethod
    def MultiPoint(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.MultiPoint", "engine.Geometry", invoke_args
        )

    @staticmethod
    def MultiLineString(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.MultiLineString", "engine.Geometry", invoke_args
        )

    @staticmethod
    def MultiPolygon(coordinates: list) -> engine.Geometry:
        if coordinates is not None and not isinstance(coordinates, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"coordinates 只支持list类型参数, 传入类型为{type(coordinates)}",
            )

        invoke_args = {
            "coordinates": coordinates,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "coordinates" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数coordinates不能为空")

        return FunctionHelper.apply(
            "GeometryConstructors.MultiPolygon", "engine.Geometry", invoke_args
        )

    def voronoiDiagram(self) -> engine.Geometry:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Geometry.voronoiDiagram", "engine.Geometry", invoke_args
        )

    def delaunayTriangulation(self) -> engine.Geometry:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Geometry.delaunayTriangulation", "engine.Geometry", invoke_args
        )

    def smooth(
        self, iterations: int, offset: [int, float], maxAngle: [int, float]
    ) -> engine.Geometry:
        if iterations is not None and not isinstance(iterations, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"iterations 只支持int类型参数, 传入类型为{type(iterations)}",
            )

        if offset is not None and not isinstance(offset, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"offset 只支持(int,float)类型参数, 传入类型为{type(offset)}",
            )

        if maxAngle is not None and not isinstance(maxAngle, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"maxAngle 只支持(int,float)类型参数, 传入类型为{type(maxAngle)}",
            )

        invoke_args = {
            "geometry": self,
            "iterations": iterations,
            "offset": offset,
            "maxAngle": maxAngle,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "iterations" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数iterations不能为空")

        if "offset" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数offset不能为空")

        if "maxAngle" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数maxAngle不能为空")

        return FunctionHelper.apply("Geometry.smooth", "engine.Geometry", invoke_args)

    def area(self) -> object:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.area", "object", invoke_args)

    def perimeter(self) -> object:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.perimeter", "object", invoke_args)

    def length(self) -> object:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.length", "object", invoke_args)

    def coordinates(self) -> object:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.coordinates", "object", invoke_args)

    def type(self) -> object:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.type", "object", invoke_args)

    def bounds(self) -> engine.Geometry:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.bounds", "engine.Geometry", invoke_args)

    def centroid(self) -> engine.Geometry:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.centroid", "engine.Geometry", invoke_args)

    def convexHull(self) -> engine.Geometry:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "Geometry.convexHull", "engine.Geometry", invoke_args
        )

    def cutLines(self, distance: list) -> engine.Geometry:
        if distance is not None and not isinstance(distance, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"distance 只支持list类型参数, 传入类型为{type(distance)}"
            )

        invoke_args = {
            "geometry": self,
            "distance": distance,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "distance" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数distance不能为空")

        return FunctionHelper.apply("Geometry.cutLines", "engine.Geometry", invoke_args)

    def simplify(self, tolerance: [int, float]) -> engine.Geometry:
        if tolerance is not None and not isinstance(tolerance, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"tolerance 只支持(int,float)类型参数, 传入类型为{type(tolerance)}",
            )

        invoke_args = {
            "geometry": self,
            "tolerance": tolerance,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "tolerance" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数tolerance不能为空")

        return FunctionHelper.apply("Geometry.simplify", "engine.Geometry", invoke_args)

    def transform(self, proj: str) -> engine.Geometry:
        if proj is not None and not isinstance(proj, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"proj 只支持str类型参数, 传入类型为{type(proj)}"
            )

        invoke_args = {
            "geometry": self,
            "proj": proj,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "proj" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数proj不能为空")

        return FunctionHelper.apply(
            "Geometry.transform", "engine.Geometry", invoke_args
        )

    def distance(self, right: engine.Geometry) -> object:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Geometry.distance", "object", invoke_args)

    def withinDistance(self, right: engine.Geometry, distance: [int, float]) -> object:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        if distance is not None and not isinstance(distance, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"distance 只支持(int,float)类型参数, 传入类型为{type(distance)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
            "distance": distance,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        if "distance" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数distance不能为空")

        return FunctionHelper.apply("Geometry.withinDistance", "object", invoke_args)

    def containedIn(self, right: engine.Geometry) -> object:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Geometry.containedIn", "object", invoke_args)

    def contains(self, right: engine.Geometry) -> object:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Geometry.contains", "object", invoke_args)

    def difference(self, right: engine.Geometry) -> engine.Geometry:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply(
            "Geometry.difference", "engine.Geometry", invoke_args
        )

    def disjoint(self, right: engine.Geometry) -> object:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Geometry.disjoint", "object", invoke_args)

    def dissolve(self) -> engine.Geometry:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.dissolve", "engine.Geometry", invoke_args)

    def intersection(self, right: engine.Geometry) -> engine.Geometry:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply(
            "Geometry.intersection", "engine.Geometry", invoke_args
        )

    def intersects(self, right: engine.Geometry) -> object:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Geometry.intersects", "object", invoke_args)

    def symmetricDifference(self, right: engine.Geometry) -> engine.Geometry:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply(
            "Geometry.symmetricDifference", "engine.Geometry", invoke_args
        )

    def union(self, right: engine.Geometry) -> engine.Geometry:
        if right is not None and not isinstance(right, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"right 只支持engine.Geometry类型参数, 传入类型为{type(right)}",
            )

        invoke_args = {
            "left": self,
            "right": right,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "right" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数right不能为空")

        return FunctionHelper.apply("Geometry.union", "engine.Geometry", invoke_args)

    def buffer(self, distance: [int, float]) -> engine.Geometry:
        if distance is not None and not isinstance(distance, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"distance 只支持(int,float)类型参数, 传入类型为{type(distance)}",
            )

        invoke_args = {
            "geometry": self,
            "distance": distance,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "distance" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数distance不能为空")

        return FunctionHelper.apply("Geometry.buffer", "engine.Geometry", invoke_args)

    def geometries(self) -> object:
        invoke_args = {
            "geometry": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Geometry.geometries", "object", invoke_args)

    def coveringGrid(self, interval: [int, float]) -> engine.FeatureCollection:
        if interval is not None and not isinstance(interval, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"interval 只支持(int,float)类型参数, 传入类型为{type(interval)}",
            )

        invoke_args = {
            "geometry": self,
            "interval": interval,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "interval" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数interval不能为空")

        return FunctionHelper.apply(
            "Geometry.coveringGrid", "engine.FeatureCollection", invoke_args
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: UDFImages.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0fUDFImages.proto\x12\x08protocol\"9\n\x07Wrapper\x12\x1e\n\x05image\x18\x01 \x03(\x0b\x32\x0f.protocol.Image\x12\x0e\n\x06params\x18\x02 \x01(\t\"y\n\x05Image\x12/\n\x08\x62\x61ndData\x18\x01 \x03(\x0b\x32\x1d.protocol.Image.BandDataEntry\x1a?\n\rBandDataEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1d\n\x05value\x18\x02 \x01(\x0b\x32\x0e.protocol.Band:\x02\x38\x01\"[\n\x04\x42\x61nd\x12\x10\n\x08tileData\x18\x01 \x01(\x0c\x12\x10\n\x08\x64\x61taType\x18\x02 \x01(\t\x12\x13\n\x0bnodataValue\x18\x03 \x01(\t\x12\x0c\n\x04\x63ols\x18\x04 \x01(\x05\x12\x0c\n\x04rows\x18\x05 \x01(\x05\x42\x33\n\x1c\x63om.alibaba.aie.engine.modelB\tUDFImagesZ\x08protocolb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'UDFImages_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\034com.alibaba.aie.engine.modelB\tUDFImagesZ\010protocol'
  _IMAGE_BANDDATAENTRY._options = None
  _IMAGE_BANDDATAENTRY._serialized_options = b'8\001'
  _globals['_WRAPPER']._serialized_start=29
  _globals['_WRAPPER']._serialized_end=86
  _globals['_IMAGE']._serialized_start=88
  _globals['_IMAGE']._serialized_end=209
  _globals['_IMAGE_BANDDATAENTRY']._serialized_start=146
  _globals['_IMAGE_BANDDATAENTRY']._serialized_end=209
  _globals['_BAND']._serialized_start=211
  _globals['_BAND']._serialized_end=302
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations


from traitlets import HasTraits, Union, List, Unicode, Float, validate, Integer

from ipyleaflet import projections, WMSPostLayer, WMSLayer
from aiearth.core.client.endpoints import Endpoints
from aiearth.core.auth import Authenticate
from aiearth.core.error import AIEError, AIEErrorCode
from collections.abc import Iterable
from aiearth import engine
from aiearth.engine.map import aie_color


class ImageVisParams(HasTraits):
    bands = Union(
        [List(Integer()), Integer(), Unicode(), List(Unicode())], default_value=None, allow_none=True
    ).tag(desc="list of band names (or single band) to be mapped to RGB")
    min = Union([List(Float()), Float()], default_value=None, allow_none=True).tag(
        desc="min values (or one per band) to map onto 00."
    )
    max = Union([List(Float()), Float()], default_value=None, allow_none=True).tag(
        desc="max values (or one per band) to map onto FF."
    )
    gain = Union([List(Float()), Float()], default_value=None, allow_none=True).tag(
        desc="gain values (or one per band) to map onto 00-FF."
    )
    bias = Union([List(Float()), Float()], default_value=None, allow_none=True).tag(
        desc="bias values (or one per band) to map onto FF."
    )
    gamma = Union([List(Float()), Float()], default_value=None, allow_none=True).tag(
        desc="gamma correction factors (or one per band)."
    )
    palette = List(Unicode(), minlen=2, default_value=None, allow_none=True).tag(
        desc="color palette for single band render"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def symbolize_params(self):
        to_8bit_method = "ColorRamp" if self.palette is not None else "MinMaxStretch"
        to_8bit_params = (
            self.color_map_params()
            if self.palette is not None
            else self.stretch_params()
        )
        return {
            "symbolizeMethodNames": [to_8bit_method],
            "symbolizeMethodParams": {to_8bit_method: to_8bit_params},
        }

    def color_map_params(self):
        min_value = (
            (0 - self.bias[0] / self.gain[0]) if self.min is None else self.min[0]
        )
        max_value = (
            (255 - self.bias[0] / self.gain[0]) if self.max is None else self.max[0]
        )
        color_value = min_value
        color_value_step = (max_value - min_value) / (len(self.palette) - 1)
        color_map = []
        for color in self.palette:
            color_map.append({"value": color_value, "color": color})
            color_value += color_value_step
        return {"renderMode": 1, "colorMap": color_map}

    def stretch_params(self):
        bands = len(self.gain) if self.min is None else len(self.min)
        min_values = (
            [0 - self.bias[i] / self.gain[i] for i in range(bands)]
            if self.min is None
            else self.min
        )
        max_values = (
            [255 - self.bias[i] / self.gain[i] for i in range(bands)]
            if self.max is None
            else self.max
        )
        return {
            "fromMinMaxPerBand": [[min_values[i], max_values[i]] for i in range(bands)]
        }

    @validate("bands")
    def _validate_bands(self, proposal):
        bands = proposal["value"]
        if bands is None:
            return bands
        elif not isinstance(bands, list):
            return [bands]
        elif len(bands) != 1 and len(bands) != 3:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "选择的波段数必须为 1 或 3")
        else:
            return bands

    @validate("min", "max", "gain", "bias", "gamma")
    def _validate_stretch_params(self, proposal):
        name = proposal["trait"].name
        if name in ["min", "max"] and (self.min is None or self.max is None):
            raise AIEError(AIEErrorCode.ARGS_ERROR, "必须同时设置 min 和 max！")
        elif name in ["gain", "bias"]:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "暂时不支持 gain 和 bias！")
        elif self.min is not None and self.gain is not None:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "不能同时设置 min/max 和 gain/bias！")
        values = proposal["value"]
        if not isinstance(values, list):
            return [values] * 1 if self.bands is None else len(self.bands)
        elif len(values) == 1:
            return values * 1 if self.bands is None else len(self.bands)
        else:
            return values

    @validate("palette")
    def _validate_palette(self, proposal):
        palette_colors = proposal["value"]
        if self.min is None and self.gain is None:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "palette 使用需要设置 min/max 或 gain/bias")
        for color in palette_colors:
            if len(color) != 7:
                raise AIEError(AIEErrorCode.ARGS_ERROR, "颜色 '%s' 格式不满足要求, 例子: '#FFFFFF'" % color)
            try:
                int(color[1:], 16)
            except ValueError:
                raise AIEError(AIEErrorCode.ARGS_ERROR, "颜色 '%s' 格式不满足要求, 例子: '#FFFFFF'" % color)
        return palette_colors


class AIEImageLayer(WMSLayer):
    token = Unicode().tag(sync=True, o=True)
    aie_map_id = Unicode().tag(sync=True, o=True)

    def __init__(self, name: str, aie_object: engine.Image, vis_params: dict, **kwargs):
        if "bounds" not in kwargs:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "当前需要显式指定 aie_object 的 bounds")
        if "crs" in kwargs.keys():
            kwargs.pop("crs")
            print("warning. 暂不支持坐标系设置，默认为EPSG4326")

        bounds = kwargs["bounds"]
        bounds = [[bounds[1], bounds[0]], [bounds[3], bounds[2]]]
        kwargs.pop("bounds")

        if 'palette' in vis_params.keys() \
                and isinstance(vis_params['palette'], Iterable):
            # valid only one band provided if palette exsits
            if 'bands' in vis_params \
                    and not isinstance(vis_params['bands'], str) \
                    and isinstance(vis_params['bands'], Iterable) \
                    and len(vis_params['bands']) > 1:
                raise ValueError('调色盘模式(Palette)仅支持一个波段的渲染')
            vis_params['palette'] = [aie_color.translate_color(color) for color in vis_params['palette']]
        image_vis_params = ImageVisParams(**vis_params)
        if image_vis_params.bands is not None:
            aie_object = aie_object.select(image_vis_params.bands)
        map_id = aie_object.getMapId(vis_params)
        new_kwargs = {
            "name": name,
            "url": Endpoints.MAP,
            "bounds": bounds,
            "layers": "async_aie_map_raster",
            "crs": projections.EPSG4326,
            "aie_map_id": 'Maps_mapId_' + map_id["mapId"],
            "camel_case": False,
            "attribution": "AIE Engine",
            "token": Authenticate.getCurrentUserToken(),
        }
        new_kwargs.update(kwargs)
        super().__init__(**new_kwargs)


class AIEGeometryLayer(WMSLayer):
    token = Unicode().tag(sync=True, o=True)
    aie_map_id = Unicode().tag(sync=True, o=True)

    def __init__(self, name: str, aie_object: engine.Geometry, vis_params: dict, **kwargs):
        if "bounds" not in kwargs:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "当前需要显式指定 aie_object 的 bounds")
        bounds = kwargs["bounds"]
        bounds = [[bounds[1], bounds[0]], [bounds[3], bounds[2]]]
        kwargs.pop("bounds")
        if 'color' in vis_params and isinstance(vis_params['color'], str):
            vis_params['color'] = aie_color.translate_color(vis_params['color'])
        if 'styles' in vis_params and isinstance(vis_params['styles'], Iterable):
            for style in vis_params['styles']:
                if isinstance(style, dict):
                    style['color'] = aie_color.translate_color(style['color'])
        if "color" not in vis_params:
            vis_params["color"] = "#000000"
        map_id = aie_object.getMapId(vis_params)
        new_kwargs = {
            "name": name,
            "url": Endpoints.MAP_VECTOR,
            "bounds": bounds,
            "layers": "mpg",
            "crs": projections.EPSG4326,
            "token": Authenticate.getCurrentUserToken(),
            "aie_map_id": 'Maps_mapId_' + map_id["mapId"],
            "attribution": "AIE Engine",
            "camel_case": False
        }
        new_kwargs.update(kwargs)
        super().__init__(**new_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class FeatureCollection(engine.Collection):
    def __init__(self, args) -> engine.FeatureCollection:
        if isinstance(args, str):
            invoke_args = {"id": args}
            super(FeatureCollection, self).__init__(
                "FeatureCollection.load", invoke_args
            )
        elif isinstance(args, engine.Feature):
            args = [args]
            invoke_args = {"features": args}
            super(FeatureCollection, self).__init__(
                "FeatureCollection.fromFeatures", invoke_args
            )
        elif isinstance(args, (list, tuple)):
            features = [engine.Feature(i) for i in args]
            invoke_args = {"features": features}
            super(FeatureCollection, self).__init__(
                "FeatureCollection.fromFeatures", invoke_args
            )
        elif isinstance(args, engine.List):
            super(FeatureCollection, self).__init__(
                "FeatureCollection.fromFeatures", {"features": args}
            )
        else:
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"args 只支持str|engine.Feature|list|engine.List类型参数, 传入类型为{type(args)}",
            )

    def getMapId(self, vis_params):
        if vis_params is not None and not isinstance(vis_params, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"vis_params 只支持dict类型参数, 传入类型为{type(vis_params)}",
            )
        return engine.client.Maps.getMapId(self, vis_params)

    def getCenter(self) -> tuple:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            center = ((bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2)
            return center
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Center失败. bbox: {bbox}")

    def getBounds(self) -> list:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            bounds = [bbox[0], bbox[1], bbox[2], bbox[3]]
            return bounds
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Bounds失败. bbox: {bbox}")

    def elementType(self):
        return engine.Feature

    def propertyNames(self) -> object:
        invoke_args = {
            "element": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "FeatureCollection.propertyNames", "object", invoke_args
        )

    def aggregate_max(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_max", "object", invoke_args
        )

    def aggregate_min(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_min", "object", invoke_args
        )

    def aggregate_mean(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_mean", "object", invoke_args
        )

    def aggregate_count(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_count", "object", invoke_args
        )

    def aggregate_sum(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_sum", "object", invoke_args
        )

    def aggregate_stats(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_stats", "object", invoke_args
        )

    def aggregate_array(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_array", "object", invoke_args
        )

    def aggregate_product(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_product", "object", invoke_args
        )

    def aggregate_sample_sd(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_sample_sd", "object", invoke_args
        )

    def aggregate_sample_var(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_sample_var", "object", invoke_args
        )

    def aggregate_total_sd(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_total_sd", "object", invoke_args
        )

    def aggregate_total_var(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_total_var", "object", invoke_args
        )

    def aggregate_histogram(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_histogram", "object", invoke_args
        )

    def aggregate_count_distinct(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_count_distinct", "object", invoke_args
        )

    def aggregate_first(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.aggregate_first", "object", invoke_args
        )

    def reduceColumns(
        self, reducer: engine.Reducer, selectors: list, weightSelectors: list = None
    ) -> object:
        if reducer is not None and not isinstance(reducer, engine.Reducer):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"reducer 只支持engine.Reducer类型参数, 传入类型为{type(reducer)}",
            )

        if selectors is not None and not isinstance(selectors, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"selectors 只支持list类型参数, 传入类型为{type(selectors)}",
            )

        if weightSelectors is not None and not isinstance(weightSelectors, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"weightSelectors 只支持list类型参数, 传入类型为{type(weightSelectors)}",
            )

        invoke_args = {
            "collection": self,
            "reducer": reducer,
            "selectors": selectors,
            "weightSelectors": weightSelectors,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "reducer" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数reducer不能为空")

        if "selectors" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数selectors不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.reduceColumns", "object", invoke_args
        )

    def union(self, maxError: int = None) -> engine.FeatureCollection:
        if maxError is not None and not isinstance(maxError, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"maxError 只支持int类型参数, 传入类型为{type(maxError)}"
            )

        invoke_args = {
            "collection": self,
            "maxError": maxError,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "FeatureCollection.union", "engine.FeatureCollection", invoke_args
        )

    def distinct(self, properties: list) -> engine.FeatureCollection:
        if properties is not None and not isinstance(properties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"properties 只支持list类型参数, 传入类型为{type(properties)}",
            )

        invoke_args = {
            "collection": self,
            "properties": properties,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "properties" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数properties不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.distinct", "engine.FeatureCollection", invoke_args
        )

    def map(
        self, baseAlgorithm: types.FunctionType
    ) -> Union[engine.ImageCollection, engine.FeatureCollection]:
        node = super(FeatureCollection, self).map(baseAlgorithm)
        baseAlgorithm_return_type = type(node.invoke_args["baseAlgorithm"].body)
        if engine.Image == baseAlgorithm_return_type:
            return FunctionHelper.cast(node, "engine.ImageCollection")
        else:
            return FunctionHelper.cast(node, "engine.FeatureCollection")

    def merge(self, collection2: engine.FeatureCollection) -> engine.FeatureCollection:
        if collection2 is not None and not isinstance(
            collection2, engine.FeatureCollection
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"collection2 只支持engine.FeatureCollection类型参数, 传入类型为{type(collection2)}",
            )

        invoke_args = {
            "collection1": self,
            "collection2": collection2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "collection2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数collection2不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.merge", "engine.FeatureCollection", invoke_args
        )

    def randomColumn(
        self, columnName: str = "random", seed: int = 0, distribution: str = "uniform"
    ) -> engine.FeatureCollection:
        if columnName is not None and not isinstance(columnName, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"columnName 只支持str类型参数, 传入类型为{type(columnName)}",
            )

        if seed is not None and not isinstance(seed, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"seed 只支持int类型参数, 传入类型为{type(seed)}"
            )

        if distribution is not None and not isinstance(distribution, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"distribution 只支持str类型参数, 传入类型为{type(distribution)}",
            )

        invoke_args = {
            "collection": self,
            "columnName": columnName,
            "seed": seed,
            "distribution": distribution,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "FeatureCollection.randomColumn", "engine.FeatureCollection", invoke_args
        )

    @staticmethod
    def randomPoints(
        region: engine.Geometry, points: int = 1000, seed: int = 0, maxError: int = None
    ) -> engine.FeatureCollection:
        if region is not None and not isinstance(region, engine.Geometry):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"region 只支持engine.Geometry类型参数, 传入类型为{type(region)}",
            )

        if points is not None and not isinstance(points, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"points 只支持int类型参数, 传入类型为{type(points)}"
            )

        if seed is not None and not isinstance(seed, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"seed 只支持int类型参数, 传入类型为{type(seed)}"
            )

        if maxError is not None and not isinstance(maxError, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"maxError 只支持int类型参数, 传入类型为{type(maxError)}"
            )

        invoke_args = {
            "region": region,
            "points": points,
            "seed": seed,
            "maxError": maxError,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "region" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数region不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.randomPoints", "engine.FeatureCollection", invoke_args
        )

    def select(
        self,
        propertySelectors: list,
        newProperties: list = None,
        retainGeometry: bool = True,
    ) -> engine.FeatureCollection:
        if propertySelectors is not None and not isinstance(propertySelectors, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"propertySelectors 只支持list类型参数, 传入类型为{type(propertySelectors)}",
            )

        if newProperties is not None and not isinstance(newProperties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"newProperties 只支持list类型参数, 传入类型为{type(newProperties)}",
            )

        if retainGeometry is not None and not isinstance(retainGeometry, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"retainGeometry 只支持bool类型参数, 传入类型为{type(retainGeometry)}",
            )

        invoke_args = {
            "featurecollection": self,
            "propertySelectors": propertySelectors,
            "newProperties": newProperties,
            "retainGeometry": retainGeometry,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "propertySelectors" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数propertySelectors不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.select", "engine.FeatureCollection", invoke_args
        )

    def filter(self, filter: engine.Filter) -> engine.FeatureCollection:
        node = super(FeatureCollection, self).filter(filter)
        return FunctionHelper.cast(node, "engine.FeatureCollection")

    def filterBounds(
        self, geometry: Union[engine.Geometry, engine.Feature, engine.FeatureCollection]
    ) -> engine.FeatureCollection:
        node = super(FeatureCollection, self).filterBounds(geometry)
        return FunctionHelper.cast(node, "engine.FeatureCollection")

    def first(self) -> engine.Feature:
        node = super(FeatureCollection, self).first()
        return FunctionHelper.cast(node, "engine.Feature")

    def limit(
        self, max: [int, float], property: str = None, ascending: bool = None
    ) -> engine.FeatureCollection:
        node = super(FeatureCollection, self).limit(max, property=None, ascending=None)
        return FunctionHelper.cast(node, "engine.FeatureCollection")

    def geometry(self) -> engine.Geometry:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "FeatureCollection.geometry", "engine.Geometry", invoke_args
        )

    def toList(self, count: int, offset: int = 0) -> object:
        if count is not None and not isinstance(count, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"count 只支持int类型参数, 传入类型为{type(count)}"
            )

        if offset is not None and not isinstance(offset, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"offset 只支持int类型参数, 传入类型为{type(offset)}"
            )

        invoke_args = {
            "collection": self,
            "count": count,
            "offset": offset,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "count" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数count不能为空")

        return FunctionHelper.apply("FeatureCollection.toList", "object", invoke_args)

    def sort(self, property: str, ascending: bool = True) -> engine.FeatureCollection:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        if ascending is not None and not isinstance(ascending, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"ascending 只支持bool类型参数, 传入类型为{type(ascending)}",
            )

        invoke_args = {
            "collection": self,
            "property": property,
            "ascending": ascending,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.sort", "engine.FeatureCollection", invoke_args
        )

    def classify(
        self, classifier: engine.Classifier, outputName: str = "classification"
    ) -> engine.FeatureCollection:
        if classifier is not None and not isinstance(classifier, engine.Classifier):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"classifier 只支持engine.Classifier类型参数, 传入类型为{type(classifier)}",
            )

        if outputName is not None and not isinstance(outputName, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"outputName 只支持str类型参数, 传入类型为{type(outputName)}",
            )

        invoke_args = {
            "input": self,
            "classifier": classifier,
            "outputName": outputName,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "classifier" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数classifier不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.classify", "engine.FeatureCollection", invoke_args
        )

    def errorMatrix(self, actual: str, predicted: str) -> engine.ConfusionMatrix:
        if actual is not None and not isinstance(actual, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"actual 只支持str类型参数, 传入类型为{type(actual)}"
            )

        if predicted is not None and not isinstance(predicted, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"predicted 只支持str类型参数, 传入类型为{type(predicted)}"
            )

        invoke_args = {
            "input": self,
            "actual": actual,
            "predicted": predicted,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "actual" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数actual不能为空")

        if "predicted" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数predicted不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.errorMatrix", "engine.ConfusionMatrix", invoke_args
        )

    def renamePropertyNames(self, var_args: dict) -> engine.FeatureCollection:
        if var_args is not None and not isinstance(var_args, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"var_args 只支持dict类型参数, 传入类型为{type(var_args)}"
            )

        invoke_args = {
            "input": self,
            "var_args": var_args,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "var_args" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数var_args不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.renamePropertyNames",
            "engine.FeatureCollection",
            invoke_args,
        )

    def makeArray(
        self, properties: list, name: str = "array"
    ) -> engine.FeatureCollection:
        if properties is not None and not isinstance(properties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"properties 只支持list类型参数, 传入类型为{type(properties)}",
            )

        if name is not None and not isinstance(name, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"name 只支持str类型参数, 传入类型为{type(name)}"
            )

        invoke_args = {
            "input": self,
            "properties": properties,
            "name": name,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "properties" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数properties不能为空")

        return FunctionHelper.apply(
            "FeatureCollection.makeArray", "engine.FeatureCollection", invoke_args
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import yaml
import os
import jinja2
import re
from collections import namedtuple
import subprocess
import sys
from jinja2 import Environment, FileSystemLoader, Template
from template.plugin_impl import IMPL

try:
    # use faster C loader if available
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def parse_native_yaml(path):
    with open(path, "r", encoding='UTF-8') as f:
        return yaml.load(f, Loader=Loader)


class Shell(object):
    @staticmethod
    def run(shell):
        try:
            command = ['bash', '-c', shell]
            process = subprocess.Popen(
                command, stdout=sys.stdout, stderr=sys.stderr)
            err = process.wait()
        except subprocess.CalledProcessError as e:
            err = e.returncode
        return err


def jinja2_render(tpl_file, params, render_file):
    loader = os.path.dirname(tpl_file)
    tpl_basename = os.path.basename(tpl_file)
    env = Environment(loader=FileSystemLoader(loader))
    tpl = env.get_template(tpl_basename)

    with open(render_file, "wb") as fout:
        render_content = tpl.render(render=params)
        fout.write(render_content.encode("utf-8"))


def to_snake(name):
    return re.sub(r"([a-z])([A-Z])", r"\1_\2", name).lower()


def gen_check_input_arg_template(arg_type, arg):
    template = Template(
        r"""
        if {{arg}} is not None and not isinstance({{arg}}, {{arg_type}}):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"{{arg}} 只支持{{arg_type}}类型参数, 传入类型为{type({{arg}})}")
    """
    )
    return template.render(arg=arg, arg_type=arg_type)


TYPE_DEF = dict()


def decl_type(arg_type, returns_type=False):
    arg_types = arg_type.split('|')
    if len(arg_types) == 0:
        raise Exception(f"arg_type 不能为空")

    if len(arg_types) == 1:
        if arg_types[0] in TYPE_DEF:
            if returns_type and arg_types[0] in ("int", "float", "bool", "str", "list", "tuple", "dict", "object"):
                return ("object", "object")
            elif arg_types[0] == "float" and not returns_type:
                return ("[int,float]", "(int,float)")
            else:
                return (TYPE_DEF[arg_types[0]], TYPE_DEF[arg_types[0]])
        else:
            raise Exception(f"arg_type: {arg_type} not support")

    union_named_type = "Union["
    parenthesis_named_type = "("
    for i in range(len(arg_types)):
        if arg_types[i] in TYPE_DEF:
            if returns_type and arg_types[i] in ("int", "float", "bool", "str", "list", "tuple", "dict", "object"):
                union_named_type = union_named_type + "object"
                parenthesis_named_type = parenthesis_named_type + "object"
            elif arg_types[i] == "float" and not returns_type:
                union_named_type = union_named_type + "int,float"
                parenthesis_named_type = parenthesis_named_type + "int,float"
            else:
                union_named_type = union_named_type + TYPE_DEF[arg_types[i]]
                parenthesis_named_type = parenthesis_named_type + \
                    TYPE_DEF[arg_types[i]]
        else:
            raise Exception(
                f"arg_type: {arg_type}. type {arg_types[i]} not support")

        if i != len(arg_types) - 1:
            union_named_type = union_named_type + ","
            parenthesis_named_type = parenthesis_named_type + ","
    union_named_type = union_named_type + "]"
    parenthesis_named_type = parenthesis_named_type + ")"
    return (union_named_type, parenthesis_named_type)


def gen_input_args(args, sig_type):
    if sig_type == "__init__" or sig_type == "self":
        input_args = "self"
    elif sig_type == "staticmethod":
        input_args = ""
    elif sig_type == "override":
        input_args = "self"
    elif sig_type == "abstractmethod":
        input_args = "self"
    else:
        raise Exception(f"sig_type {sig_type} not support")

    for arg in args:
        if arg.bind_arg is None or arg.bind_arg != "self":
            input_arg = arg.name + ":" + decl_type(arg.type)[0]
            if input_args != "":
                input_args = input_args + ", "
            input_args = input_args + input_arg

            if arg.default is not None:
                input_args = input_args + " = " + arg.default

    return input_args


def gen_override_args(args):
    input_args = ""
    for arg in args:
        if arg.bind_arg is None or arg.bind_arg != "self":
            input_arg = arg.name
            if input_args != "":
                input_args = input_args + ", "
            input_args = input_args + input_arg

            if arg.default is not None:
                input_args = input_args + " = " + arg.default

    return input_args


def gen_invoke_args(args):
    invoke_args = dict()
    for arg in args:
        if len(arg) < 3 or arg[2] != "self":
            invoke_args[arg[1]] = arg[1]
        else:
            invoke_args[arg[1]] = "self"
    return invoke_args


def convert_arg(func_arg):
    Arg = namedtuple(
        "Arg", "type name bind_arg default"
    )

    arg_type, arg = func_arg.split(' ')
    if arg_type == "" or arg == "":
        raise Exception(f"func_arg: {func_arg} 缺少arg_type、arg")
    start_colon = arg.find(':')
    start_equal = arg.find('=')
    if start_colon != -1 and start_equal != -1:
        if start_colon > start_equal:
            raise Exception(f'{func_arg}, 参数:应该在=前, eg: Image mask:self=null')

    arg_name_end = len(arg)
    if start_colon != -1:
        arg_name_end = start_colon
    elif start_equal != -1:
        arg_name_end = start_equal
    arg_name = arg[:arg_name_end]

    bind_arg = None
    if start_colon != -1:
        bind_arg_end = len(arg)
        if start_equal != -1:
            bind_arg_end = start_equal
        bind_arg = arg[start_colon+1:bind_arg_end]

    default = None
    if start_equal != -1:
        default = arg[start_equal+1:]

    return Arg(arg_type, arg_name, bind_arg, default)


def gen_check_invoke_arg_template(arg):
    template = Template(
        r"""
        if "{{arg}}" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数{{arg}}不能为空")
    """
    )
    return template.render(arg=arg)

def parse_function(kclass, signature):
    Func = namedtuple(
        "Func", "impl annotation func_def func_invoke sig_type input_args check_input_args invoke_args override_args check_invoke_args returns"
    )

    if "plugin_impl" in signature:
        if not signature["plugin_impl"] in IMPL:
            raise Exception(
                f'plugin_impl: {signature["plugin_impl"]} 在 template/plugin_impl 中没有实现')
        impl = IMPL[signature["plugin_impl"]]
        return Func(impl, None, None, None, None, None, None, None, None, None, None)

    func_splited = signature["func"].split("->")
    if len(func_splited) != 2:
        raise Exception(f"func signature {signature['func']} 缺少返回值")
    func, returns = func_splited[0], func_splited[1].strip()
    if returns == "":
        raise Exception(f"func {func} 缺少返回值")
    returns_type = decl_type(returns, True)[0]
    args_start = func.find("(")
    args_end = func.rfind(")")
    args_str = func[args_start + 1: args_end]
    func_invoke = func[:args_start].strip()

    args = [convert_arg(arg.strip())
            for arg in args_str.split(",") if args_str != ""]

    sig_type = "self"
    if "type" in signature:
        sig_type = signature["type"].strip()
    func_def = func_invoke
    if sig_type == "__init__":
        func_def = "__init__"
    if "func_def_alias" in signature:
        func_def = signature["func_def_alias"]

    input_args = gen_input_args(args, sig_type)
    check_input_args = [
        gen_check_input_arg_template(decl_type(arg.type)[1], arg.name)
        for arg in args if arg.bind_arg is None or arg.bind_arg != "self"
    ]
    invoke_args = gen_invoke_args(args)

    check_invoke_args = [
        gen_check_invoke_arg_template(arg.name)
        for arg in args if (arg.bind_arg is None or arg.bind_arg != "self") and arg.default is None
    ]

    func_invoke = kclass + '.' + func_invoke
    if "func_invoke_alias" in signature:
        func_invoke = signature["func_invoke_alias"]

    annotation = None
    if sig_type == "__init__" or sig_type == "self" or sig_type == "override":
        pass
    elif sig_type == "staticmethod":
        annotation = "@staticmethod"
    elif sig_type == "abstractmethod":
        annotation == "@abc.abstractmethod"
    else:
        raise Exception(f"sig_type {sig_type} not support")

    override_args = None
    if sig_type == "override":
        override_args = gen_override_args(args)

    return Func(
        None,
        annotation,
        func_def,
        func_invoke,
        sig_type,
        input_args,
        check_input_args,
        invoke_args,
        override_args,
        check_invoke_args,
        returns_type,
    )


def reformat(file):
    shell = f'black {file}'
    err = Shell.run(shell)
    if err != 0:
        raise Exception(f"file: {file} reformat error")


def generate_aie_object_code(source_path, declarations):
    for name, signatures in declarations.items():
        params = dict()
        parent_kclass_start = name.find("(")
        parent_kclass_end = name.find(")")
        if parent_kclass_start != -1 or parent_kclass_end != -1:
            parent_kclass = name[parent_kclass_start + 1: parent_kclass_end]
            if parent_kclass == "":
                raise Exception(f"{name} 父类解析为空")
            params["parent_kclass"] = decl_type(parent_kclass, True)[1]
            kclass = name[:parent_kclass_start]
        else:
            params["parent_kclass"] = "FunctionNode"
            kclass = name

        params["kclass"] = kclass
        params["funcs"] = [parse_function(kclass,
                                          signature) for signature in signatures]

        tpl_file = source_path + "/template/aie_object.template"
        # render_file = source_path + "/aie_object/" + \
        #     to_snake(kclass) + "_gen.py"
        render_file = source_path + "/aie_object/" + \
            to_snake(kclass) + ".py"
        jinja2_render(tpl_file, params, render_file)
        reformat(render_file)


def generate_module_init_code(source_path, declarations):
    import_decls = list()
    for name, signatures in declarations.items():
        parent_kclass_start = name.find("(")
        if parent_kclass_start != -1:
            kclass = name[:parent_kclass_start]
        else:
            kclass = name
        # module_file = to_snake(kclass) + "_gen"
        module_file = to_snake(kclass)
        import_decls.append("from aiearth.engine.aie_object." +
                            module_file + " import " + kclass)

    params = {"import_decls": import_decls}

    tpl_file = source_path + "/template/__init__.template"
    render_file = source_path + "/__init__.py"
    jinja2_render(tpl_file, params, render_file)
    reformat(render_file)


def main():
    source_path = os.path.dirname(os.path.abspath(__file__))
    global TYPE_DEF
    TYPE_DEF = parse_native_yaml(source_path + "/type_def.yaml")
    declarations = parse_native_yaml(source_path + "/functions.yaml")
    print(TYPE_DEF)
    print(declarations)
    generate_aie_object_code(source_path, declarations)
    generate_module_init_code(source_path, declarations)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from .function_node import FunctionNode
from . import aie_object
from aiearth.core.error import AIEError, AIEErrorCode
from aiearth import engine


class FunctionHelper(object):
    @classmethod
    def cast(cls, node, kclass):
        if kclass == "engine.Image":
            node.__class__ = engine.Image
        elif kclass == "engine.Collection":
            node.__class__ = engine.Collection
        elif kclass == "engine.ImageCollection":
            node.__class__ = engine.ImageCollection
        elif kclass == "engine.Geometry":
            node.__class__ = engine.Geometry
        elif kclass == "engine.Feature":
            node.__class__ = engine.Feature
        elif kclass == "engine.FeatureCollection":
            node.__class__ = engine.FeatureCollection
        elif kclass == "engine.Filter":
            node.__class__ = engine.Filter
        elif kclass == "engine.Reducer":
            node.__class__ = engine.Reducer
        elif kclass == "engine.Terrain":
            node.__class__ = engine.Terrain
        elif kclass == "engine.Kernel":
            node.__class__ = engine.Kernel
        elif kclass == "engine.Classifier":
            node.__class__ = engine.Classifier
        elif kclass == "engine.ConfusionMatrix":
            node.__class__ = engine.ConfusionMatrix
        elif kclass == 'engine.List':
            node.__class__ = engine.List
        elif kclass == 'engine.Number':
            node.__class__ = engine.Number
        elif kclass in ("int", "float", "bool", "str", "list", "tuple", "dict", "object"):
            pass
        else:
            raise AIEError(AIEErrorCode.DEFAULT_INTERNAL_ERROR,
                           "", "FunctionHelper::cast kclass " + kclass + " not support")
        return node

    @classmethod
    def apply(cls, name, returns, args):
        node = FunctionNode(name, args)
        return cls.cast(node, returns)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class Terrain(FunctionNode):
    @staticmethod
    def aspect(input: engine.Image) -> engine.Image:
        if input is not None and not isinstance(input, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"input 只支持engine.Image类型参数, 传入类型为{type(input)}",
            )

        invoke_args = {
            "input": input,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "input" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数input不能为空")

        return FunctionHelper.apply("Terrain.aspect", "engine.Image", invoke_args)

    @staticmethod
    def slope(input: engine.Image) -> engine.Image:
        if input is not None and not isinstance(input, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"input 只支持engine.Image类型参数, 传入类型为{type(input)}",
            )

        invoke_args = {
            "input": input,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "input" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数input不能为空")

        return FunctionHelper.apply("Terrain.slope", "engine.Image", invoke_args)

    @staticmethod
    def hillshade(
        input: engine.Image, azimuth: [int, float] = 270, elevation: [int, float] = 45
    ) -> engine.Image:
        if input is not None and not isinstance(input, engine.Image):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"input 只支持engine.Image类型参数, 传入类型为{type(input)}",
            )

        if azimuth is not None and not isinstance(azimuth, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"azimuth 只支持(int,float)类型参数, 传入类型为{type(azimuth)}",
            )

        if elevation is not None and not isinstance(elevation, (int, float)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"elevation 只支持(int,float)类型参数, 传入类型为{type(elevation)}",
            )

        invoke_args = {
            "input": input,
            "azimuth": azimuth,
            "elevation": elevation,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "input" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数input不能为空")

        return FunctionHelper.apply("Terrain.hillshade", "engine.Image", invoke_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class Projection(FunctionNode):
    def __init__(self, string: str) -> engine.Projection:
        if string is not None and not isinstance(string, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"string 只支持str类型参数, 传入类型为{type(string)}"
            )

        invoke_args = {
            "string": string,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "string" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数string不能为空")

        super(Projection, self).__init__("Projection.constructors", invoke_args)

    def crs(self) -> object:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Projection.crs", "object", invoke_args)

    def wkt(self) -> object:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Projection.wkt", "object", invoke_args)

    def proj(self) -> object:
        invoke_args = {
            "input": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("Projection.proj", "object", invoke_args)

    def coordinateTransform(self, src: engine.Projection, xs: list, ys: list) -> object:
        if src is not None and not isinstance(src, engine.Projection):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"src 只支持engine.Projection类型参数, 传入类型为{type(src)}",
            )

        if xs is not None and not isinstance(xs, list):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"xs 只支持list类型参数, 传入类型为{type(xs)}")

        if ys is not None and not isinstance(ys, list):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"ys 只支持list类型参数, 传入类型为{type(ys)}")

        invoke_args = {
            "input": self,
            "src": src,
            "xs": xs,
            "ys": ys,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "src" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数src不能为空")

        if "xs" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数xs不能为空")

        if "ys" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数ys不能为空")

        return FunctionHelper.apply(
            "Projection.coordinateTransform", "object", invoke_args
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import annotations

import abc
import inspect
from typing import Union
from types import FunctionType

from aiearth import engine
from aiearth.engine.variable_node import VariableNode
from aiearth.engine.function_node import FunctionNode
from aiearth.engine.customfunction_node import CustomFunctionNode
from aiearth.engine.function_helper import FunctionHelper
from aiearth.core.error.aie_error import AIEError, AIEErrorCode


class ImageCollection(engine.Collection):
    def __init__(self, args) -> engine.ImageCollection:
        if isinstance(args, str):
            invoke_args = {"id": args}
            super(ImageCollection, self).__init__("ImageCollection.load", invoke_args)
        elif isinstance(args, engine.Image):
            args = [args]
            invoke_args = {"images": args}
            super(ImageCollection, self).__init__(
                "ImageCollection.fromImages", invoke_args
            )
        elif isinstance(args, (list, tuple)):
            images = [engine.Image(i) for i in args]
            invoke_args = {"images": images}
            super(ImageCollection, self).__init__(
                "ImageCollection.fromImages", invoke_args
            )
        elif isinstance(args, engine.List):
            super(ImageCollection, self).__init__(
                "ImageCollection.fromImages", {"images": args}
            )
        else:
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"args 只支持str|engine.Image|list|engine.List类型参数, 传入类型为{type(args)}",
            )

    def elementType(self):
        return engine.Image

    def getMapId(self, vis_params):
        if vis_params is not None and not isinstance(vis_params, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"vis_params 只支持dict类型参数, 传入类型为{type(vis_params)}",
            )
        return engine.client.Maps.getMapId(self, vis_params)

    def mosaic(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "ImageCollection.mosaic", "engine.Image", invoke_args
        )

    def qualityMosaic(self, qualityBand: str) -> engine.Image:
        if qualityBand is not None and not isinstance(qualityBand, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"qualityBand 只支持str类型参数, 传入类型为{type(qualityBand)}",
            )

        invoke_args = {
            "collection": self,
            "qualityBand": qualityBand,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "qualityBand" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数qualityBand不能为空")

        return FunctionHelper.apply(
            "ImageCollection.qualityMosaic", "engine.Image", invoke_args
        )

    def getCenter(self) -> tuple:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            center = ((bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2)
            return center
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Center失败. bbox: {bbox}")

    def getBounds(self) -> list:
        bbox = engine.client.InteractiveSession.getBounds(self)
        if bbox is not None and isinstance(bbox, list):
            bounds = [bbox[0], bbox[1], bbox[2], bbox[3]]
            return bounds
        raise AIEError(AIEErrorCode.ARGS_ERROR, f"获取Bounds失败. bbox: {bbox}")

    def toList(self, count: int, offset: int = 0) -> object:
        if count is not None and not isinstance(count, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"count 只支持int类型参数, 传入类型为{type(count)}"
            )

        if offset is not None and not isinstance(offset, int):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"offset 只支持int类型参数, 传入类型为{type(offset)}"
            )

        invoke_args = {
            "collection": self,
            "count": count,
            "offset": offset,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "count" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数count不能为空")

        return FunctionHelper.apply("ImageCollection.toList", "object", invoke_args)

    def filter(self, filter: Union[str, engine.Filter]) -> engine.ImageCollection:
        node = super(ImageCollection, self).filter(filter)
        return FunctionHelper.cast(node, "engine.ImageCollection")

    def filterBounds(
        self, geometry: Union[engine.Geometry, engine.Feature, engine.FeatureCollection]
    ) -> engine.ImageCollection:
        node = super(ImageCollection, self).filterBounds(geometry)
        return FunctionHelper.cast(node, "engine.ImageCollection")

    def filterDate(self, start: str, end: str) -> engine.ImageCollection:
        if start is not None and not isinstance(start, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"start 只支持str类型参数, 传入类型为{type(start)}"
            )

        if end is not None and not isinstance(end, str):
            raise AIEError(AIEErrorCode.ARGS_ERROR, f"end 只支持str类型参数, 传入类型为{type(end)}")

        invoke_args = {
            "collection": self,
            "start": start,
            "end": end,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "start" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数start不能为空")

        if "end" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数end不能为空")

        return FunctionHelper.apply(
            "Collection.filterDate", "engine.ImageCollection", invoke_args
        )

    def first(self) -> engine.Image:
        node = super(ImageCollection, self).first()
        return FunctionHelper.cast(node, "engine.Image")

    def limit(
        self, limit: int, property: str = None, ascending: bool = True
    ) -> engine.ImageCollection:
        node = super(ImageCollection, self).limit(limit, property=None, ascending=True)
        return FunctionHelper.cast(node, "engine.ImageCollection")

    def reduce(self, reducer: engine.Reducer) -> engine.Image:
        if reducer is not None and not isinstance(reducer, engine.Reducer):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"reducer 只支持engine.Reducer类型参数, 传入类型为{type(reducer)}",
            )

        invoke_args = {
            "collection": self,
            "reducer": reducer,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "reducer" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数reducer不能为空")

        return FunctionHelper.apply(
            "ImageCollection.reduce", "engine.Image", invoke_args
        )

    def And(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.and", "engine.Image", invoke_args)

    def count(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.count", "engine.Image", invoke_args)

    def max(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.max", "engine.Image", invoke_args)

    def mean(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.mean", "engine.Image", invoke_args)

    def median(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.median", "engine.Image", invoke_args)

    def min(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.min", "engine.Image", invoke_args)

    def mode(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.mode", "engine.Image", invoke_args)

    def Or(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.or", "engine.Image", invoke_args)

    def product(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.product", "engine.Image", invoke_args)

    def sum(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply("reduce.sum", "engine.Image", invoke_args)

    def map(
        self, baseAlgorithm: types.FunctionType
    ) -> Union[engine.ImageCollection, engine.FeatureCollection]:
        node = super(ImageCollection, self).map(baseAlgorithm)
        baseAlgorithm_return_type = type(node.invoke_args["baseAlgorithm"].body)
        if engine.Feature == baseAlgorithm_return_type:
            return FunctionHelper.cast(node, "engine.FeatureCollection")
        else:
            return FunctionHelper.cast(node, "engine.ImageCollection")

    def select(self, selectors) -> engine.ImageCollection:
        return self.map(lambda image: image.select(selectors))

    def sort(self, property: str, ascending: bool = True) -> engine.ImageCollection:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        if ascending is not None and not isinstance(ascending, bool):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"ascending 只支持bool类型参数, 传入类型为{type(ascending)}",
            )

        invoke_args = {
            "collection": self,
            "property": property,
            "ascending": ascending,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.sort", "engine.ImageCollection", invoke_args
        )

    def toBands(self) -> engine.Image:
        invoke_args = {
            "collection": self,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        return FunctionHelper.apply(
            "ImageCollection.toBands", "engine.Image", invoke_args
        )

    @staticmethod
    def fromImages(images: Union[list, engine.List]) -> engine.ImageCollection:
        if images is not None and not isinstance(images, (list, engine.List)):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"images 只支持(list,engine.List)类型参数, 传入类型为{type(images)}",
            )

        invoke_args = {
            "images": images,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "images" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数images不能为空")

        return FunctionHelper.apply(
            "ImageCollection.fromImages", "engine.ImageCollection", invoke_args
        )

    def merge(self, collection2: engine.ImageCollection) -> engine.ImageCollection:
        if collection2 is not None and not isinstance(
            collection2, engine.ImageCollection
        ):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"collection2 只支持engine.ImageCollection类型参数, 传入类型为{type(collection2)}",
            )

        invoke_args = {
            "input": self,
            "collection2": collection2,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "collection2" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数collection2不能为空")

        return FunctionHelper.apply(
            "ImageCollection.merge", "engine.ImageCollection", invoke_args
        )

    def cast(self, bandTypes: dict) -> engine.ImageCollection:
        if bandTypes is not None and not isinstance(bandTypes, dict):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"bandTypes 只支持dict类型参数, 传入类型为{type(bandTypes)}",
            )

        invoke_args = {
            "input": self,
            "bandTypes": bandTypes,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "bandTypes" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数bandTypes不能为空")

        return FunctionHelper.apply(
            "ImageCollection.cast", "engine.ImageCollection", invoke_args
        )

    def aggregate_max(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_max", "object", invoke_args
        )

    def aggregate_min(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_min", "object", invoke_args
        )

    def aggregate_mean(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_mean", "object", invoke_args
        )

    def aggregate_count(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_count", "object", invoke_args
        )

    def aggregate_sum(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_sum", "object", invoke_args
        )

    def aggregate_stats(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_stats", "object", invoke_args
        )

    def aggregate_array(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_array", "object", invoke_args
        )

    def aggregate_product(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_product", "object", invoke_args
        )

    def aggregate_sample_sd(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_sample_sd", "object", invoke_args
        )

    def aggregate_sample_var(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_sample_var", "object", invoke_args
        )

    def aggregate_total_sd(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_total_sd", "object", invoke_args
        )

    def aggregate_total_var(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_total_var", "object", invoke_args
        )

    def aggregate_histogram(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_histogram", "object", invoke_args
        )

    def aggregate_count_distinct(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_count_distinct", "object", invoke_args
        )

    def aggregate_first(self, property: str) -> object:
        if property is not None and not isinstance(property, str):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR, f"property 只支持str类型参数, 传入类型为{type(property)}"
            )

        invoke_args = {
            "collection": self,
            "property": property,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "property" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数property不能为空")

        return FunctionHelper.apply(
            "ImageCollection.aggregate_first", "object", invoke_args
        )

    def reduceToImage(self, properties: list, reducer: engine.Reducer) -> engine.Image:
        if properties is not None and not isinstance(properties, list):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"properties 只支持list类型参数, 传入类型为{type(properties)}",
            )

        if reducer is not None and not isinstance(reducer, engine.Reducer):
            raise AIEError(
                AIEErrorCode.ARGS_ERROR,
                f"reducer 只支持engine.Reducer类型参数, 传入类型为{type(reducer)}",
            )

        invoke_args = {
            "collection": self,
            "properties": properties,
            "reducer": reducer,
        }

        invoke_args = {k: v for k, v in invoke_args.items() if v is not None}

        if "properties" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数properties不能为空")

        if "reducer" not in invoke_args:
            raise AIEError(AIEErrorCode.ARGS_ERROR, "参数reducer不能为空")

        return FunctionHelper.apply(
            "ImageCollection.reduceToImage", "engine.Image", invoke_args
        )
